"""Helpers for working with ESI."""

import datetime as dt
import logging
import random
from time import sleep
from typing import Optional

import requests

from django.utils.timezone import now

from app_utils.logging import LoggerAddTag

from . import __title__, __version__
from ._app_settings import (
    APPUTILS_ESI_DAILY_DOWNTIME_END,
    APPUTILS_ESI_DAILY_DOWNTIME_START,
    APPUTILS_ESI_ERROR_LIMIT_THRESHOLD,
)

logger = LoggerAddTag(logging.getLogger(__name__), __title__)


class EsiStatusException(Exception):
    """EsiStatus base exception."""

    def __init__(self, message):
        super().__init__()
        self.message = message


class EsiOffline(EsiStatusException):
    """ESI is offline error."""

    def __init__(self):
        """:meta private:"""
        super().__init__("ESI appears to be offline.")


class EsiDailyDowntime(EsiOffline):
    """ESI is offline due to daily downtime."""

    def __init__(self):
        """:meta private:"""
        super().__init__()
        self.message = "Assuming ESI is offline during it's daily downtime."


class EsiErrorLimitExceeded(EsiStatusException):
    """ESI error limit exceeded error."""

    def __init__(self, retry_in: float = 60) -> None:
        """:meta private:"""
        super().__init__("The ESI error limit has been exceeded.")
        self._retry_in = float(retry_in)

    @property
    def retry_in(self) -> float:
        """Time until next error window in seconds."""
        return self._retry_in


class EsiStatus:
    """Current status of ESI (immutable)."""

    MAX_JITTER = 20

    def __init__(
        self,
        is_online: bool,
        error_limit_remain: Optional[int] = None,
        error_limit_reset: Optional[int] = None,
        is_daily_downtime: bool = False,
    ) -> None:
        self._is_online = bool(is_online)
        self._is_daily_downtime = is_daily_downtime
        if error_limit_remain is None or error_limit_reset is None:
            self._error_limit_remain = None
            self._error_limit_reset = None
        else:
            self._error_limit_remain = int(error_limit_remain)
            self._error_limit_reset = int(error_limit_reset)

    @property
    def is_ok(self) -> bool:
        """True if ESI is online and below error limit, else False."""
        return self.is_online and not self.is_error_limit_exceeded

    @property
    def is_daily_downtime(self) -> bool:
        """True if status was created during daily downtime time frame."""
        return self._is_daily_downtime

    @property
    def is_online(self) -> bool:
        """True if ESI is online, else False."""
        return self._is_online

    @property
    def error_limit_remain(self) -> Optional[int]:
        """Amount of remaining errors in current window."""
        return self._error_limit_remain

    @property
    def error_limit_reset(self) -> Optional[int]:
        """Seconds until current error window resets."""
        return self._error_limit_reset

    @property
    def is_error_limit_exceeded(self) -> bool:
        """True when remain is below the threshold, else False.

        Will also return False if remain/reset are not defined
        """
        return bool(
            self.error_limit_remain
            and self.error_limit_reset
            and self.error_limit_remain <= APPUTILS_ESI_ERROR_LIMIT_THRESHOLD
        )

    def error_limit_reset_w_jitter(self, max_jitter: Optional[int] = None) -> int:
        """Calc seconds to retry in order to reach next error window incl. jitter."""
        if self.error_limit_reset is None:
            return 0
        if not max_jitter or max_jitter < 1:
            max_jitter = self.MAX_JITTER
        return self.error_limit_reset + int(random.uniform(1, max_jitter))

    def raise_for_status(self):
        """Raise an exception if ESI if offline or the error limit is exceeded."""
        if not self.is_online:
            if self.is_daily_downtime:
                raise EsiDailyDowntime()
            raise EsiOffline()
        if self.is_error_limit_exceeded:
            raise EsiErrorLimitExceeded(retry_in=self.error_limit_reset_w_jitter())


def fetch_esi_status(ignore_daily_downtime: bool = False) -> EsiStatus:
    """Determine the current ESI status.

    Args:
        ignore_daily_downtime: When True will always make a request to ESI \
            even during the daily downtime
    """
    is_daily_downtime = _is_daily_downtime()
    if not ignore_daily_downtime and is_daily_downtime:
        return EsiStatus(is_online=False, is_daily_downtime=True)
    try:
        response = _request_esi_status()
    except (requests.exceptions.Timeout, requests.exceptions.ConnectionError):
        logger.warning("Network error when trying to call ESI", exc_info=True)
        return EsiStatus(is_online=False, is_daily_downtime=is_daily_downtime)
    if not response.ok:
        is_online = False
    else:
        try:
            is_online = not response.json().get("vip")
        except ValueError:
            is_online = False
    try:
        remain = int(response.headers.get("X-Esi-Error-Limit-Remain"))  # type: ignore
        reset = int(response.headers.get("X-Esi-Error-Limit-Reset"))  # type: ignore
    except TypeError:
        logger.warning("Failed to parse HTTP headers: %s", response.headers)
        return EsiStatus(is_online=is_online, is_daily_downtime=is_daily_downtime)
    logger.debug(
        "ESI status: is_online: %s, error_limit_remain = %s, error_limit_reset = %s",
        is_online,
        remain,
        reset,
    )
    return EsiStatus(
        is_online=is_online,
        error_limit_remain=remain,
        error_limit_reset=reset,
        is_daily_downtime=is_daily_downtime,
    )


def _is_daily_downtime() -> bool:
    """Determine if we currently are in the daily downtime period."""
    downtime_start = _calc_downtime(APPUTILS_ESI_DAILY_DOWNTIME_START)
    downtime_end = _calc_downtime(APPUTILS_ESI_DAILY_DOWNTIME_END)
    return now() >= downtime_start and now() <= downtime_end


def _calc_downtime(hours_float: float) -> dt.datetime:
    hour, minute = _convert_float_hours(hours_float)
    return now().replace(hour=hour, minute=minute)


def _convert_float_hours(hours_float: float) -> tuple:
    """Convert float hours into int hours and int minutes for datetime."""
    hours = int(hours_float)
    minutes = int((hours_float - hours) * 60)
    return hours, minutes


def _request_esi_status() -> requests.Response:
    """Fetch current status from ESI. Retry on common HTTP errors."""
    max_retries = 3
    retry_count = 0
    while True:
        response = requests.get(
            "https://esi.evetech.net/latest/status/",
            timeout=(5, 30),
            headers={"User-Agent": f"{__package__};{__version__}"},
        )
        if response.status_code not in {
            502,  # HTTPBadGateway
            503,  # HTTPServiceUnavailable
            504,  # HTTPGatewayTimeout
        }:
            break

        retry_count += 1
        if retry_count > max_retries:
            break

        logger.warning(
            "HTTP status code %s - Try %s/%s",
            response.status_code,
            retry_count,
            max_retries,
        )

        wait_secs = 0.1 * (random.uniform(2, 4) ** (retry_count - 1))
        sleep(wait_secs)

    return response


def retry_task_if_esi_is_down(self):
    """Retry current celery task if ESI is not online or error threshold is exceeded.

    This function has to be called from inside a celery task!

    Args:
        self: Current celery task from `@shared_task(bind=True)`
    """
    try:
        fetch_esi_status().raise_for_status()
    except EsiOffline as ex:
        countdown = (5 + int(random.uniform(1, 10))) * 60
        logger.warning(
            "ESI appears to be offline. Trying again in %d seconds.", countdown
        )
        raise self.retry(countdown=countdown) from ex
    except EsiErrorLimitExceeded as ex:
        logger.warning(
            "ESI error limit threshold reached. Trying again in %s seconds", ex.retry_in
        )
        raise self.retry(countdown=ex.retry_in) from ex



"""Utilities for the admin site."""

from collections import Counter

from django.contrib import admin
from django.db.models import Count


class FieldFilterCountsMemory(admin.SimpleListFilter):
    """Filter by field and show counts for admin site.

    Counts are calculated in memory.
    """

    field_name = ""  # field to filter by

    def lookups(self, request, model_admin: admin.ModelAdmin):
        """:meta private:"""
        field_in_rows = model_admin.get_queryset(request).values_list(
            self.field_name, flat=True
        )
        field_counts = Counter(field_in_rows)
        result = [
            (field, f"{field} ({count:,})") for field, count in field_counts.items()
        ]
        return sorted(result, key=lambda obj: obj[0])

    def queryset(self, request, queryset):
        """:meta private:"""
        if self.value():
            params = {self.field_name: self.value()}
            return queryset.filter(**params)


class FieldFilterCountsDb(admin.SimpleListFilter):
    """Filter by field and show counts for admin site.

    Counts are calculated by the database.
    """

    field_name = ""  # field to filter by

    def lookups(self, request, model_admin: admin.ModelAdmin):
        """:meta private:"""
        qs = model_admin.get_queryset(request)
        field = qs.model._meta.get_field(self.field_name)
        if not field.choices:
            qs = qs.exclude(**{self.field_name: ""})
        field_counts = (
            qs.values(self.field_name)
            .annotate(num_words=Count(self.field_name))
            .order_by(self.field_name)
        )
        if field.choices:
            field_counts = self.__map_choices_field(field, field_counts)
            result = [
                (
                    obj[self.field_name][0],
                    f'{obj[self.field_name][1]} ({obj["num_words"]:,})',
                )
                for obj in field_counts
            ]
        else:
            result = [
                (obj[self.field_name], f'{obj[self.field_name]} ({obj["num_words"]:,})')
                for obj in field_counts
            ]
        return result

    def __map_choices_field(self, field, field_counts):
        """Map choices field values to corresponding labels and keep values."""
        mapper = {obj[0]: obj[1] for obj in field.choices}
        field_counts = [
            {
                self.field_name: (obj[self.field_name], mapper[obj[self.field_name]]),
                "num_words": obj["num_words"],
            }
            for obj in field_counts
        ]
        return field_counts

    def queryset(self, request, queryset):
        """:meta private:"""
        if self.value():
            params = {self.field_name: self.value()}
            return queryset.filter(**params)



"""JSON related utilities."""

import datetime as dt
import json
from typing import Any

from pytz import timezone


class JSONDateTimeDecoder(json.JSONDecoder):
    """Decoder for the standard json library to decode JSON into datetime.
    To be used together with ``JSONDateTimeEncoder``.

    Example:

        .. code-block:: python

            message = json.loads(message_json, cls=JSONDateTimeDecoder)

    """

    def __init__(self, *args, **kwargs) -> None:
        """
        :meta private:
        """
        json.JSONDecoder.__init__(
            self, object_hook=self._dict_to_object, *args, **kwargs
        )

    def _dict_to_object(self, dct: dict) -> object:
        if "__type__" not in dct:
            return dct

        type_str = dct.pop("__type__")
        zone, _ = dct.pop("tz")
        if zone:
            dct["tzinfo"] = timezone(zone)
        try:
            date_obj = dt.datetime(**dct)
            return date_obj
        except (ValueError, TypeError):
            dct["__type__"] = type_str
            return dct


class JSONDateTimeEncoder(json.JSONEncoder):
    """Encoder for the standard json library to encode datetime into JSON.
    To be used together with ``JSONDateTimeDecoder``.

    Works with naive and aware datetimes, but only with UTC timezone.

    Example:

        .. code-block:: python

            message_json = json.dumps(message, cls=JSONDateTimeEncoder)

    """

    def default(self, o: Any) -> Any:
        """
        :meta private:
        """
        if isinstance(o, dt.datetime):
            tz_name = o.tzinfo.tzname(o) if o.tzinfo else None
            if tz_offset := o.utcoffset():
                tz_total_seconds = tz_offset.total_seconds()
            else:
                tz_total_seconds = None
            timezone_info = tz_name, tz_total_seconds
            return {
                "__type__": "datetime",
                "year": o.year,
                "month": o.month,
                "day": o.day,
                "hour": o.hour,
                "minute": o.minute,
                "second": o.second,
                "microsecond": o.microsecond,
                "tz": timezone_info,
            }

        return json.JSONEncoder.default(self, o)



"""This module provides factories for generating test objects from Django and AA Models.

Important: You need to add the dependency ``factory_boy`` to your test environment.
"""

from typing import Generic, TypeVar

import factory
import factory.fuzzy

from django.contrib.auth import get_user_model
from django.db.models import Max

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.tests.auth_utils import AuthUtils

from .testing import add_character_to_user

T = TypeVar("T")
User = get_user_model()


class BaseMetaFactory(Generic[T], factory.base.FactoryMetaClass):
    def __call__(cls, *args, **kwargs) -> T:
        return super().__call__(*args, **kwargs)


# django


class UserFactory(factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[User]):
    """Generate a User object.

    Args:
        permissions: List of permission names (optional),
            e.g. ``["moonmining.basic_access"]``
    """

    class Meta:
        model = User
        django_get_or_create = ("username",)
        exclude = ("_generated_name",)

    _generated_name = factory.Faker("name")
    username = factory.LazyAttribute(lambda obj: obj._generated_name.replace(" ", "_"))
    first_name = factory.LazyAttribute(lambda obj: obj._generated_name.split(" ")[0])
    last_name = factory.LazyAttribute(lambda obj: obj._generated_name.split(" ")[1])
    email = factory.LazyAttribute(
        lambda obj: f"{obj.first_name.lower()}.{obj.last_name.lower()}@example.com"
    )

    @factory.post_generation
    def permissions(obj, create, extracted, **kwargs):
        """Set default permissions. Overwrite with `permissions=["app.perm1"]`."""
        if not create:
            return
        permissions = extracted or []
        for permission_name in permissions:
            AuthUtils.add_permission_to_user_by_name(permission_name, obj)

    @classmethod
    def _after_postgeneration(cls, obj, create, results=None):
        """Reset permission cache to force an update."""
        super()._after_postgeneration(obj, create, results)
        if hasattr(obj, "_perm_cache"):
            del obj._perm_cache
        if hasattr(obj, "_user_perm_cache"):
            del obj._user_perm_cache


# auth


class EveAllianceInfoFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveAllianceInfo]
):
    """Generate an EveAllianceInfo object."""

    class Meta:
        model = EveAllianceInfo
        django_get_or_create = ("alliance_id", "alliance_name")

    alliance_name = factory.Faker("catch_phrase")
    alliance_ticker = factory.LazyAttribute(lambda obj: obj.alliance_name[:4].upper())
    executor_corp_id = 0

    @factory.lazy_attribute
    def alliance_id(self):
        last_id = (
            EveAllianceInfo.objects.aggregate(Max("alliance_id"))["alliance_id__max"]
            or 99_000_000
        )
        return last_id + 1


class EveCorporationInfoFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveCorporationInfo]
):
    """Generate an EveCorporationInfo object."""

    class Meta:
        model = EveCorporationInfo
        django_get_or_create = ("corporation_id", "corporation_name")

    corporation_name = factory.Faker("catch_phrase")
    corporation_ticker = factory.LazyAttribute(
        lambda obj: obj.corporation_name[:4].upper()
    )
    member_count = factory.fuzzy.FuzzyInteger(1000)

    @factory.lazy_attribute
    def corporation_id(self):
        last_id = (
            EveCorporationInfo.objects.aggregate(Max("corporation_id"))[
                "corporation_id__max"
            ]
            or 98_000_000
        )
        return last_id + 1

    @factory.post_generation
    def create_alliance(obj, create, extracted, **kwargs):
        if not create or extracted is False or obj.alliance:
            return
        obj.alliance = EveAllianceInfoFactory(executor_corp_id=obj.corporation_id)


class EveCharacterFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[EveCharacter]
):
    """Generate an EveCharacter object."""

    class Meta:
        model = EveCharacter
        django_get_or_create = ("character_id", "character_name")
        exclude = ("corporation",)

    character_name = factory.Faker("name")
    corporation = factory.SubFactory(EveCorporationInfoFactory)
    corporation_id = factory.LazyAttribute(lambda obj: obj.corporation.corporation_id)
    corporation_name = factory.LazyAttribute(
        lambda obj: obj.corporation.corporation_name
    )
    corporation_ticker = factory.LazyAttribute(
        lambda obj: obj.corporation.corporation_ticker
    )

    @factory.lazy_attribute
    def character_id(self):
        last_id = (
            EveCharacter.objects.aggregate(Max("character_id"))["character_id__max"]
            or 90_000_000
        )
        return last_id + 1

    @factory.lazy_attribute
    def alliance_id(self):
        return (
            self.corporation.alliance.alliance_id if self.corporation.alliance else None
        )

    @factory.lazy_attribute
    def alliance_name(self):
        return (
            self.corporation.alliance.alliance_name if self.corporation.alliance else ""
        )

    @factory.lazy_attribute
    def alliance_ticker(self):
        return (
            self.corporation.alliance.alliance_ticker
            if self.corporation.alliance
            else ""
        )


class UserMainFactory(UserFactory):
    """Generate a User object with main character.

    Args:
        main_character__character: EveCharacter object to be used as main (optional)
        main_character__scopes: List of ESI scope names (optional),
            e.g. ``["esi-characters.read_contacts.v1"]``
    """

    @factory.post_generation
    def main_character(obj, create, _extracted, **kwargs):
        if not create:
            return
        if "character" in kwargs:  # TODO: maybe use extracted directly here?
            character = kwargs["character"]
        else:
            character_name = f"{obj.first_name} {obj.last_name}"
            character = EveCharacterFactory(character_name=character_name)

        scopes = kwargs.get("scopes", None)
        add_character_to_user(
            user=obj, character=character, is_main=True, scopes=scopes
        )



"""Utilities for logging."""

import logging


class LoggerAddTag(logging.LoggerAdapter):
    """adds a custom prefix to the given logger

    Example:

        .. code-block:: python

            import logging
            from app_utils.logging import LoggerAddTag

            logger = LoggerAddTag(logging.getLogger(__name__), __package__)

    """

    def __init__(self, my_logger, prefix):
        """:meta private:"""
        super().__init__(my_logger, {})
        self.prefix = prefix

    def process(self, msg, kwargs):
        """:meta private:"""
        return (
            "[%s] %s" % (self.prefix, msg),  # pylint: disable=consider-using-f-string
            kwargs,
        )


logger = LoggerAddTag(logging.getLogger(__name__), __package__)


def make_logger_prefix(tag: str):
    """creates a function to add logger prefixes, which returns tag when used empty"""
    return lambda text="": f'{tag}{(": " + text) if text else ""}'



"""Utilities and helpers for Python scripts."""

import inspect
import os
import sys
from pathlib import Path
from typing import Optional


def start_django(
    max_hops=10,
    django_project_name="myauth",
    settings_module="myauth.settings.local",
    silent=False,
) -> None:
    """Start the current Django project.

    This function encapsulates the boilerplate code needed to start the current Django
    project in a normal Python script.

    It will also try to detect the path to the current Django project.
    If it can not be found, the function will exit with code 1.

    Args:
        - max_hops: Max number of hops up on the main path to check
        - django_project_name: Name of the Django project
        - settings_module: Qualified name of the settings module in the Django project
        - silent: When True will not produce any output

    Here is an example how to use this function in a script:

    .. code-block:: python

        from app_utils.scripts import start_django

        start_django()

        def main():
            # put Django imports here
            ...

        if __name__ == "__main__":
            main()

    '''
    """

    # calc path of caller script
    previous_frame = inspect.currentframe().f_back
    traceback = inspect.getframeinfo(previous_frame)
    caller_path = Path(traceback.filename).parent

    django_path = _find_django_path(caller_path, max_hops, django_project_name)
    if not django_path:
        if not silent:
            print(
                f"FATAL: Could not find the {django_project_name} "
                f"folder within {max_hops} hops in the path above {caller_path}"
            )
        sys.exit(1)

    sys.path.insert(0, str(django_path))
    if not silent:
        print(f"Starting Django ({django_path})")
    import django

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", settings_module)
    django.setup()


def _find_django_path(
    start_path: Path, max_search_depth: int, django_project_name: str
) -> Optional[Path]:
    parent_path = start_path
    for _ in range(max_search_depth):
        django_path = parent_path / django_project_name
        if django_path.exists() and (django_path / "manage.py").exists():
            return django_path

        parent_path = parent_path.parent

    return None



"""Extending the Django utilities."""

# pylint: disable = unused-import

from typing import Optional

from django.apps import apps
from django.contrib.auth.models import Permission, User
from django.db import models
from django.utils.html import format_html

from .app_settings import clean_setting  # noqa: F401 - for backwards compatibility only
from .urls import static_file_absolute_url


def app_labels() -> set:
    """returns set of all current app labels"""
    return set(apps.app_configs.keys())


def users_with_permission(
    permission: Permission, include_superusers=True
) -> models.QuerySet:
    """returns queryset of users that have the given Django permission

    Args:
        permission: required permission
        include_superusers: whether superusers are included in the returned list
    """
    users_qs = (
        permission.user_set.all()
        | User.objects.filter(
            groups__in=list(permission.group_set.values_list("pk", flat=True))
        )
        | User.objects.select_related("profile").filter(
            profile__state__in=list(permission.state_set.values_list("pk", flat=True))
        )
    )
    if include_superusers:
        users_qs |= User.objects.filter(is_superuser=True)
    return users_qs.distinct()


def admin_boolean_icon_html(value) -> Optional[str]:
    """Variation of the admin boolean type, which returns the HTML for creating
    the usual `True` and `False` icons.
    But returns `None` for `None`, instead of the question mark."""

    def make_html(icon_url: str, alt: str) -> str:
        return format_html(f'<img src="{icon_url}" alt="{alt}">')

    if value is True:
        icon_url = static_file_absolute_url("admin/img/icon-yes.svg")
        return make_html(icon_url, "True")

    if value is False:
        icon_url = static_file_absolute_url("admin/img/icon-no.svg")
        return make_html(icon_url, "False")

    return None



"""Utilities for making it easier to write tests."""

# pylint: disable=unused-import

import datetime as dt
import json
import logging
import os
import re
import socket
from itertools import count
from typing import Any, Iterable, List, Optional, Tuple

from django.contrib.auth.models import Group, User
from django.db import models
from django.http import HttpResponse, JsonResponse
from django.test import TestCase
from esi.models import Scope, Token

from allianceauth.authentication.models import CharacterOwnership, State
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
    EveFactionInfo,
)
from allianceauth.groupmanagement.models import AuthGroup
from allianceauth.tests.auth_utils import AuthUtils

from .datetime import dt_eveformat
from .esi_testing import BravadoOperationStub, BravadoResponseStub  # noqa: F401
from .helpers import random_string


def generate_invalid_pk(MyModel: Any) -> int:
    """return an invalid PK for the given Django model"""
    pk_max = MyModel.objects.aggregate(models.Max("pk"))["pk__max"]
    return pk_max + 1 if pk_max else 1


class SocketAccessError(Exception):
    """Error raised when a test script accesses the network"""


class NoSocketsTestCase(TestCase):
    """Variation of Django's TestCase class that prevents any network use.

    Example:

        .. code-block:: python

            class TestMyStuff(NoSocketsTestCase):
                def test_should_do_what_i_need(self):
                    ...

    """

    @classmethod
    def setUpClass(cls):
        cls.socket_original = socket.socket
        socket.socket = cls.guard
        return super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        socket.socket = cls.socket_original
        return super().tearDownClass()

    @staticmethod
    def guard(*args, **kwargs):
        """:meta_private:"""
        raise SocketAccessError("Attempted to access network")


def set_test_logger(logger_name: str, name: str) -> object:
    """set logger for current test module

    Args:
        logger: current logger object
        name: name of current module, e.g. __file__

    Returns:
        amended logger
    """

    # reconfigure logger so we get logging from tested module
    f_format = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(module)s:%(funcName)s - %(message)s"
    )
    file_name = os.path.splitext(name)[0]
    f_handler = logging.FileHandler(f"{file_name}.log", "w+")
    f_handler.setFormatter(f_format)
    my_logger = logging.getLogger(logger_name)
    my_logger.level = logging.DEBUG
    my_logger.addHandler(f_handler)
    my_logger.propagate = False
    return my_logger


def queryset_pks(queryset) -> set:
    """shortcut that returns the pks of the given queryset as set.
    Useful for comparing test results.
    """
    return set(queryset.values_list("pk", flat=True))


def response_text(response: HttpResponse) -> str:
    """Return content of a HTTP response as string."""
    return response.content.decode("utf-8")


def json_response_to_python(response: JsonResponse) -> Any:
    """Convert JSON response into Python object."""
    return json.loads(response_text(response))


def json_response_to_dict(response: JsonResponse, key="id") -> dict:
    """Convert JSON response into dict by given key."""
    return {x[key]: x for x in json_response_to_python(response)}


def multi_assert_in(items: Iterable, container: Iterable) -> bool:
    """Return True if all items are in container."""
    for item in items:
        if item not in container:
            return False

    return True


def multi_assert_not_in(items: Iterable, container: Iterable) -> bool:
    """Return True if none of the item is in container."""
    for item in items:
        if item in container:
            return False

    return True


# factories


def add_new_token(
    user: User,
    character: EveCharacter,
    scopes: Optional[List[str]] = None,
    owner_hash: Optional[str] = None,
) -> Token:
    """Generate a new token for a user based on a character."""
    return _store_as_Token(
        _generate_token(
            character_id=character.character_id,
            character_name=character.character_name,
            owner_hash=owner_hash,
            scopes=scopes,
        ),
        user,
    )


def _generate_token(
    character_id: int,
    character_name: str,
    owner_hash: Optional[str] = None,
    access_token: str = "access_token",
    refresh_token: str = "refresh_token",
    scopes: Optional[list] = None,
    timestamp_dt: Optional[dt.datetime] = None,
    expires_in: int = 1200,
) -> dict:
    """Generates the input to create a new SSO test token"""
    if timestamp_dt is None:
        timestamp_dt = dt.datetime.utcnow()
    if scopes is None:
        scopes = [
            "esi-mail.read_mail.v1",
            "esi-wallet.read_character_wallet.v1",
            "esi-universe.read_structures.v1",
        ]
    if owner_hash is None:
        owner_hash = random_string(28)
    token = {
        "access_token": access_token,
        "token_type": "Bearer",
        "expires_in": expires_in,
        "refresh_token": refresh_token,
        "timestamp": int(timestamp_dt.timestamp()),
        "CharacterID": character_id,
        "CharacterName": character_name,
        "ExpiresOn": dt_eveformat(timestamp_dt + dt.timedelta(seconds=expires_in)),
        "Scopes": " ".join(list(scopes)),
        "TokenType": "Character",
        "CharacterOwnerHash": owner_hash,
        "IntellectualProperty": "EVE",
    }
    return token


def _store_as_Token(token: dict, user: object) -> Token:
    """Stores a generated token dict as Token object for given user

    returns Token object
    """
    character_tokens = user.token_set.filter(character_id=token["CharacterID"])
    if character_tokens.exists():
        token["CharacterOwnerHash"] = character_tokens.first().character_owner_hash
    obj = Token.objects.create(
        access_token=token["access_token"],
        refresh_token=token["refresh_token"],
        user=user,
        character_id=token["CharacterID"],
        character_name=token["CharacterName"],
        token_type=token["TokenType"],
        character_owner_hash=token["CharacterOwnerHash"],
    )
    for scope_name in token["Scopes"].split(" "):
        scope, _ = Scope.objects.get_or_create(name=scope_name)
        obj.scopes.add(scope)
    return obj


def create_user_from_evecharacter(
    character_id: int,
    permissions: Optional[List[str]] = None,
    scopes: Optional[List[str]] = None,
) -> Tuple[User, CharacterOwnership]:
    """Create new allianceauth user from EveCharacter object.

    Args:
        character_id: ID of eve character
        permissions: list of permission names, e.g. `"my_app.my_permission"`
        scopes: list of scope names
    """
    auth_character = EveCharacter.objects.get(character_id=character_id)
    user = AuthUtils.create_user(auth_character.character_name.replace(" ", "_"))
    character_ownership = add_character_to_user(
        user, auth_character, is_main=True, scopes=scopes
    )
    if permissions:
        for permission_name in permissions:
            user = AuthUtils.add_permission_to_user_by_name(permission_name, user)
    return user, character_ownership


def add_character_to_user(
    user: User,
    character: EveCharacter,
    is_main: bool = False,
    scopes: Optional[List[str]] = None,
    disconnect_signals: bool = False,
) -> CharacterOwnership:
    """Generates a token for the given Eve character and makes the given user it's owner

    Args:
        user: New character owner
        character: Character to add
        is_main: Will set character as the users's main when True
        scopes: List of scopes for the token
        disconnect_signals: Will disconnect signals temporarily when True
    """
    if not scopes:
        scopes = ["publicData"]

    if disconnect_signals:
        AuthUtils.disconnect_signals()
    add_new_token(user, character, scopes)

    if is_main:
        user.profile.main_character = character
        user.profile.save()
        user.save()

    if disconnect_signals:
        AuthUtils.connect_signals()

    return CharacterOwnership.objects.get(user=user, character=character)


def add_character_to_user_2(
    user: User,
    character_id,
    character_name,
    corporation_id,
    corporation_name,
    alliance_id=None,
    alliance_name=None,
    disconnect_signals=False,
) -> EveCharacter:
    """Creates a new EVE character and makes the given user the owner"""
    defaults = {
        "character_name": str(character_name),
        "corporation_id": int(corporation_id),
        "corporation_name": str(corporation_name),
    }
    if alliance_id:
        defaults["alliance_id"] = int(alliance_id)
        defaults["alliance_name"] = str(alliance_name)

    if disconnect_signals:
        AuthUtils.disconnect_signals()
    character, _ = EveCharacter.objects.update_or_create(
        character_id=character_id, defaults=defaults
    )
    CharacterOwnership.objects.create(
        character=character, owner_hash=f"{character_id}_{character_name}", user=user
    )
    if disconnect_signals:
        AuthUtils.connect_signals()

    return character


def create_fake_user(
    character_id: int,
    character_name: str,
    corporation_id: Optional[int] = None,
    corporation_name: Optional[str] = None,
    corporation_ticker: Optional[str] = None,
    alliance_id: Optional[int] = None,
    alliance_name: Optional[str] = None,
    permissions: Optional[List[str]] = None,
) -> User:
    """Create a fake user incl. main character and (optional) permissions.

    Will use default corporation and alliance if not set.
    """
    username = re.sub(r"[^\w\d@\.\+-]", "_", character_name)
    user = AuthUtils.create_user(username)
    if not corporation_id:
        corporation_id = 2001
        corporation_name = "Wayne Technologies Inc."
        corporation_ticker = "WTE"
    if corporation_id == 2001:
        alliance_id = 3001
        alliance_name = "Wayne Enterprises"
    AuthUtils.add_main_character_2(
        user=user,
        name=character_name,
        character_id=character_id,
        corp_id=corporation_id,
        corp_name=corporation_name,
        corp_ticker=corporation_ticker,
        alliance_id=alliance_id,
        alliance_name=alliance_name,
    )
    if permissions:
        perm_objs = [AuthUtils.get_permission_by_name(perm) for perm in permissions]
        user = AuthUtils.add_permissions_to_user(perms=perm_objs, user=user)
    return user


def create_authgroup(states: Optional[Iterable[State]] = None, **kwargs) -> Group:
    """Create Group object with additional Auth related properties for tests."""
    if "name" not in kwargs:
        kwargs["name"] = f"Test Group #{next_number('authgroup')}"
    name = kwargs.pop("name")
    group = Group.objects.create(name=name)
    if states:
        group.authgroup.states.add(*states)
    if kwargs:
        AuthGroup.objects.filter(group=group).update(**kwargs)
        group.authgroup.refresh_from_db()
    return group


def create_state(
    priority: int,
    permissions: Optional[Iterable[str]] = None,
    member_characters: Optional[Iterable[EveCharacter]] = None,
    member_corporations: Optional[Iterable[EveCorporationInfo]] = None,
    member_alliances: Optional[Iterable[EveAllianceInfo]] = None,
    member_factions: Optional[Iterable[EveFactionInfo]] = None,
    **kwargs,
) -> State:
    """Create a State object for tests."""
    params = {"priority": priority, "name": f"Test State #{next_number('state_name')}"}
    params.update(kwargs)
    obj = State.objects.create(**params)
    if permissions:
        perm_objs = [AuthUtils.get_permission_by_name(perm) for perm in permissions]
        obj.permissions.add(*perm_objs)
    if member_characters:
        obj.member_characters.add(*member_characters)
    if member_corporations:
        obj.member_corporations.add(*member_corporations)
    if member_alliances:
        obj.member_alliances.add(*member_alliances)
    if member_factions:
        obj.member_factions.add(*member_factions)
    return obj


def create_eve_character(
    character_id: int, character_name: str, **kwargs
) -> EveCharacter:
    """Create an EveCharacter object for tests."""
    params = {
        "character_id": character_id,
        "character_name": character_name,
        "corporation_id": 2001,
        "corporation_name": "Wayne Technologies",
        "corporation_ticker": "WYT",
        "alliance_id": 3001,
        "alliance_name": "Wayne Enterprises",
        "alliance_ticker": "WYE",
    }
    params.update(kwargs)
    return EveCharacter.objects.create(**params)


def next_number(key=None) -> int:
    """Generate consecutive numbers. Optionally numbers are generates for given key."""
    if key is None:
        key = "_general"
    try:
        return next(next_number._counter[key])
    except AttributeError:
        next_number._counter = {}
    except KeyError:
        pass
    next_number._counter[key] = count(start=1)
    return next(next_number._counter[key])



"""Improvement of the Django message class."""

from django.contrib import messages
from django.http import HttpRequest
from django.utils.safestring import mark_safe


# pylint: disable=invalid-name
class messages_plus:
    """Improvement of default Django messages with implicit HTML support."""

    @classmethod
    def debug(cls, request: HttpRequest, message: str, *args, **kwargs) -> None:
        """Send a debug message with HTML support. Use with safe strings only!"""
        messages.debug(request, mark_safe(message), *args, **kwargs)

    @classmethod
    def info(cls, request: HttpRequest, message: str, *args, **kwargs) -> None:
        """Send an info message with HTML support. Use with safe strings only!"""
        messages.info(request, mark_safe(message), *args, **kwargs)

    @classmethod
    def success(cls, request: HttpRequest, message: str, *args, **kwargs) -> None:
        """Send a success message with HTML support. Use with safe strings only!"""
        messages.success(request, mark_safe(message), *args, **kwargs)

    @classmethod
    def warning(cls, request: HttpRequest, message: str, *args, **kwargs) -> None:
        """Send a warning message with HTML support. Use with safe strings only!"""
        messages.warning(request, mark_safe(message), *args, **kwargs)

    @classmethod
    def error(cls, request: HttpRequest, message: str, *args, **kwargs) -> None:
        """Send an error message with HTML support. Use with safe strings only!"""
        messages.error(request, mark_safe(message), *args, **kwargs)



"""Utilities for caching objects and querysets."""

import functools
import hashlib
from typing import Any, Optional

from django.conf import settings
from django.core.cache import cache
from django.db import models


class ObjectCacheMixin:
    """A mixin which adds a simple object cache to a Django manager."""

    def get_cached(
        self,
        pk,
        timeout: Optional[int] = None,
        select_related: Optional[str] = None,
    ) -> Any:
        """Return the requested object either from DB or from cache.

        Can be disabled globally through the setting ``APP_UTILS_OBJECT_CACHE_DISABLED``.

        Args:
            pk: Primary key for object to fetch
            timeout: Timeout in seconds for cache
            select_related: select_related query to be applied (if any)

        Returns:
            model instance if found

        Exceptions:
            ``Model.DoesNotExist`` if object can not be found

        Example:

        .. code-block:: python

            # adding the Mixin to the model manager
            class MyModelManager(ObjectCacheMixin, models.Manager):
                pass

            # using the cache method
            obj = MyModel.objects.get_cached(pk=42, timeout=3600)

        """
        is_cache_disabled = bool(
            getattr(settings, "APP_UTILS_OBJECT_CACHE_DISABLED", False)
        )
        if is_cache_disabled:
            value = self._fetch_object_for_cache(pk=pk, select_related=select_related)
            return value

        func = functools.partial(
            self._fetch_object_for_cache, pk=pk, select_related=select_related
        )
        return cache.get_or_set(
            self._create_object_cache_key(pk, select_related), func, timeout
        )

    def clear_cache(self, pk: int):
        """Clear cache for a potentially cached object.

        This will also clear cached variants with select_related (if any).

        Args:
            pk: Primary key for object to fetch
        """
        key_base = self._create_object_base_cache_key(pk)
        cache.delete_pattern(f"{key_base}*")

    def _create_object_base_cache_key(self, pk: int) -> str:
        model_meta = self.model._meta
        return f"{model_meta.app_label}-{model_meta.model_name}-{pk}"

    def _create_object_cache_key(
        self, pk: int, select_related: Optional[str] = None
    ) -> str:
        key = self._create_object_base_cache_key(pk)
        suffix = (
            hashlib.md5(select_related.encode("utf-8")).hexdigest()
            if select_related
            else ""
        )
        return f"{key}-{suffix}" if suffix else key

    def _fetch_object_for_cache(self, pk, select_related: Optional[str] = None):
        qs = self.select_related(select_related) if select_related else self
        return qs.get(pk=pk)


def cached_queryset(queryset: models.QuerySet, key: str, timeout: int) -> Any:
    """caches the given queryset

    Args:
        queryset: the query set to cache
        key: key to be used to reference this cache
        timeout: Timeout in seconds for cache

    Returns:
        query set

    Example:

        .. code-block:: python

            queryset = cached_queryset(
                MyModel.objects.filter(name__contains="dummy"),
                key="my_cache_key",
                timeout=3600
            )

    """
    return cache.get_or_set(key, lambda: queryset, timeout)



"""Utilities related to date and time."""

import datetime as dt

import pytz

from django.utils.translation import gettext_lazy as _

# Default format for output of datetime
DATETIME_FORMAT = "%Y-%m-%d %H:%M"


def datetime_round_hour(my_dt: dt.datetime) -> dt.datetime:
    """Rounds given datetime object to nearest hour"""
    return my_dt.replace(
        second=0, microsecond=0, minute=0, hour=my_dt.hour
    ) + dt.timedelta(hours=my_dt.minute // 30)


def dt_eveformat(my_dt: dt.datetime) -> str:
    """converts a datetime to a string in eve format
    e.g. ``2019-06-25T19:04:44``
    """
    my_dt_2 = dt.datetime(
        my_dt.year, my_dt.month, my_dt.day, my_dt.hour, my_dt.minute, my_dt.second
    )
    return my_dt_2.isoformat()


def ldap_time_2_datetime(ldap_dt: int) -> dt.datetime:
    """converts ldap time to datetime.datetime"""
    return pytz.utc.localize(
        dt.datetime.utcfromtimestamp((ldap_dt / 10000000) - 11644473600)
    )


def ldap_timedelta_2_timedelta(ldap_td: int) -> dt.timedelta:
    """converts a ldap time delta into a datetime.timedelta"""
    return dt.timedelta(microseconds=ldap_td / 10)


def timeuntil_str(duration: dt.timedelta, show_seconds=True) -> str:
    """return the duration as nicely formatted string.
    Or empty string if duration is negative.

    Format: ``[[[999y] [99m]] 99d] 99h 99m 99s``
    """
    seconds = int(duration.total_seconds())
    if seconds > 0:
        periods = [
            # Translators: Abbreviation for years
            (_("y"), 60 * 60 * 24 * 365, False, True),
            # Translators: Abbreviation for months
            (_("mt"), 60 * 60 * 24 * 30, False, True),
            # Translators: Abbreviation for days
            (_("d"), 60 * 60 * 24, False, True),
            # Translators: Abbreviation for hours
            (_("h"), 60 * 60, True, True),
            # Translators: Abbreviation for months
            (_("m"), 60, True, True),
            # Translators: Abbreviation for seconds
            (_("s"), 1, True, show_seconds),
        ]
        strings = []
        for period_name, period_seconds, period_static, show in periods:
            if seconds >= period_seconds or period_static:
                period_value, seconds = divmod(seconds, period_seconds)
                if show:
                    strings.append(f"{period_value}{period_name}")

        result = " ".join(strings)
    else:
        result = ""

    return result



"""Tools for building unit tests with django-esi."""

import inspect
from collections import defaultdict
from copy import copy
from dataclasses import dataclass
from typing import Any, Callable, List, Optional, Tuple, Union

from bravado.exception import (
    HTTPBadGateway,
    HTTPBadRequest,
    HTTPError,
    HTTPForbidden,
    HTTPGatewayTimeout,
    HTTPInternalServerError,
    HTTPNotFound,
    HTTPServiceUnavailable,
    HTTPUnauthorized,
)
from pytz import utc

from django.utils.dateparse import parse_datetime


class BravadoResponseStub:
    """Stub for IncomingResponse in bravado, e.g. for HTTPError exceptions."""

    def __init__(
        self, status_code, reason="", text="", headers=None, raw_bytes=None
    ) -> None:
        self.status_code = status_code
        self.reason = reason
        self.text = text
        self.headers = headers if headers else {}
        self.raw_bytes = raw_bytes

    def __str__(self):
        return f"{self.status_code} {self.reason}"


class BravadoOperationStub:
    """Stub to simulate the operation object return from bravado via django-esi."""

    class RequestConfig:
        """A request config for a BravadoOperationStub."""

        def __init__(self, also_return_response):
            self.also_return_response = also_return_response

    def __init__(
        self,
        data,
        headers: Optional[dict] = None,
        also_return_response: bool = False,
        status_code=200,
        reason="OK",
    ):
        self._data = data
        self._headers = headers if headers else {"X-Pages": 1}
        self._status_code = status_code
        self._reason = reason
        self.request_config = BravadoOperationStub.RequestConfig(also_return_response)

    def result(self, **kwargs):
        """Execute operation and return result."""
        if self.request_config.also_return_response:
            return [
                self._data,
                BravadoResponseStub(
                    headers=self._headers,
                    status_code=self._status_code,
                    reason=self._reason,
                ),
            ]
        return self._data

    def results(self, **kwargs):
        """Execute operation and return results incl. paging."""
        return self.result(**kwargs)


def build_http_error(http_code: int, text: Optional[str] = None) -> HTTPError:
    """Build a HTTP exception for django-esi from given http code."""
    exc_map = {
        400: HTTPBadRequest,
        401: HTTPUnauthorized,
        403: HTTPForbidden,
        404: HTTPNotFound,
        500: HTTPInternalServerError,
        502: HTTPBadGateway,
        503: HTTPServiceUnavailable,
        504: HTTPGatewayTimeout,
    }
    try:
        http_exc = exc_map[http_code]
    except KeyError:
        raise NotImplementedError(f"Unknown http code: {http_code}") from None
    if not text:
        text = "Test exception"
    return http_exc(response=BravadoResponseStub(http_code, text))


@dataclass
class EsiEndpoint:
    """Class for defining ESI endpoints used in tests with the ESI client stub.

    Args:
        category: name of ESI category
        method: name of ESI method
        primary_key: name of primary key (e.g. corporation_id) or tuple of 2 keys
        needs_token: Wether the method requires a token
        data: Data to be returned from this endpoint
        http_error_code: When provided will raise an HTTP exception with this code
        side_effect: A side effect to be triggered. Can be an exception of a function.
        Exceptions will be raised.
        Functions will be called with the args of the endpoint
        and it's result returned instead of "data".
        Return the object `SIDE_EFFECT_DEFAULT` in the function
        to return the endpoints normal data.
    """

    category: str
    method: str
    primary_key: Union[str, Tuple[str, str], None] = None
    needs_token: bool = False
    data: Union[dict, list, str, None] = None
    http_error_code: Optional[int] = None
    side_effect: Union[Callable, Exception, None] = None

    def __str__(self) -> str:
        return f"{self.category}.{self.method}"

    @property
    def requires_testdata(self) -> bool:
        """True if this endpoint requires testdata to be provide as well.

        When an endpoint is only partially defined,
        one need to also provide testdata when creating a stub.
        """
        return self.data is None and not self.http_error_code and not self.side_effect


SIDE_EFFECT_DEFAULT = object()
"""Special object that can be returned from side_effect functions to indicate
that the normal data should be returned
(instead of the result of the side_effect function)
"""


class _EsiMethod:
    """An ESI method that can be called from the ESI client."""

    def __init__(
        self, endpoint: EsiEndpoint, testdata: Optional[dict], http_error: Any = False
    ) -> None:
        self._endpoint = endpoint
        if endpoint.data is not None:
            self._testdata = endpoint.data
        elif endpoint.side_effect or endpoint.http_error_code:
            self._testdata = None
        else:
            try:
                self._testdata = testdata[self._endpoint.category][
                    self._endpoint.method
                ]
            except (KeyError, TypeError):
                text = (
                    f"{self._endpoint.category}.{self._endpoint.method}: No test data"
                )
                raise build_http_error(404, text) from None
        self._http_error = http_error

    def call(self, **kwargs):
        """Method is called."""

        if isinstance(self._http_error, bool):
            if self._http_error:
                raise build_http_error(500, "Test exception")

        else:
            if isinstance(self._http_error, int):
                raise build_http_error(self._http_error, "Test exception")

        if self._endpoint.http_error_code:
            raise build_http_error(
                self._endpoint.http_error_code, "Endpoint raised exception"
            )

        if self._endpoint.side_effect:
            if inspect.isclass(self._endpoint.side_effect) and issubclass(
                self._endpoint.side_effect, Exception
            ):
                raise self._endpoint.side_effect

            result = self._endpoint.side_effect(**kwargs)
            if result != SIDE_EFFECT_DEFAULT:
                return BravadoOperationStub(result)

        pk_value = None
        if self._endpoint.primary_key:
            if isinstance(self._endpoint.primary_key, tuple):
                for pk in self._endpoint.primary_key:
                    if pk not in kwargs:
                        raise ValueError(
                            f"{self._endpoint.category}.{self._endpoint.method}: Missing primary key: {pk}"
                        )

            elif self._endpoint.primary_key not in kwargs:
                raise ValueError(
                    f"{self._endpoint.category}.{self._endpoint.method}: Missing primary key: "
                    f"{self._endpoint.primary_key}"
                )

        if self._endpoint.needs_token:
            if "token" not in kwargs:
                raise ValueError(
                    f"{self._endpoint.category}.{self._endpoint.method} "
                    f"with pk = {self._endpoint.primary_key}: Missing token"
                )

            if not isinstance(kwargs.get("token"), str):
                raise TypeError(
                    f"{self._endpoint.category}.{self._endpoint.method} "
                    f"with pk = {self._endpoint.primary_key}: Token is not a string"
                )

        try:
            if self._endpoint.primary_key:
                if isinstance(self._endpoint.primary_key, tuple):
                    pk_value_1 = str(kwargs[self._endpoint.primary_key[0]])
                    pk_value_2 = str(kwargs[self._endpoint.primary_key[1]])
                    result = self._convert_values(
                        self._testdata[pk_value_1][pk_value_2]
                    )
                else:
                    pk_value = str(kwargs[self._endpoint.primary_key])
                    result = self._convert_values(self._testdata[pk_value])

            else:
                result = self._convert_values(self._testdata)

        except (KeyError, TypeError):
            text = (
                f"{self._endpoint.category}.{self._endpoint.method}: "
                f"No test data for {self._endpoint.primary_key} = {pk_value}"
            )
            raise build_http_error(404, text) from None

        return BravadoOperationStub(result)

    @staticmethod
    def _convert_values(data) -> Any:
        def convert_dict(item):
            if isinstance(item, dict):
                for key, value in item.items():
                    if isinstance(value, str):
                        try:
                            if my_datetime := parse_datetime(value):
                                item[key] = my_datetime.replace(tzinfo=utc)
                        except ValueError:
                            pass

        if isinstance(data, list):
            for row in data:
                convert_dict(row)
        else:
            convert_dict(data)

        return data


class EsiClientStub:
    """Stub for replacing a django-esi client in tests.

    Args:
        testdata: data to be returned from Endpoint
        endpoints: List of defined endpoints
        http_error: Set `True` to generate a http 500 error exception
            or set to a http error code to generate a specific http exception
    """

    def __init__(
        self,
        testdata: Optional[dict],
        endpoints: List[EsiEndpoint],
        http_error: Any = False,
    ) -> None:
        self._testdata = testdata
        self._http_error = http_error
        self._endpoints_def = endpoints
        for endpoint in endpoints:
            self._validate_endpoint(endpoint)
            self._add_endpoint(endpoint)

    def _validate_endpoint(self, endpoint: EsiEndpoint):
        if endpoint.requires_testdata:
            try:
                _ = self._testdata[endpoint.category][endpoint.method]
            except (KeyError, TypeError):
                raise ValueError(f"No data provided for {endpoint}") from None

    def _add_endpoint(self, endpoint: EsiEndpoint):
        if not hasattr(self, endpoint.category):
            setattr(self, endpoint.category, type(endpoint.category, (object,), {}))
        my_category = getattr(self, endpoint.category)
        if not hasattr(my_category, endpoint.method):
            setattr(
                my_category,
                endpoint.method,
                _EsiMethod(
                    endpoint=endpoint,
                    testdata=self._testdata,
                    http_error=self._http_error,
                ).call,
            )
        else:
            raise ValueError(f"Endpoint for {endpoint} already defined!")

    def replace_endpoints(self, new_endpoints: List[EsiEndpoint]) -> "EsiClientStub":
        """Replace given endpoint.

        Args:
            new_endpoint: List of new endpoints

        Raises:
            ValueError: When trying to replace an non existing endpoint

        Returns:
            New stub instance with replaced endpoints
        """
        _endpoints = copy(self._endpoints_def)
        _endpoints_mapped = defaultdict(dict)
        for endpoint in _endpoints:
            _endpoints_mapped[endpoint.category][endpoint.method] = endpoint
        for new_ep in new_endpoints:
            try:
                endpoint = _endpoints_mapped[new_ep.category][new_ep.method]
            except KeyError:
                raise ValueError(f"No matching endpoint for {new_ep}") from None
            _endpoints.remove(endpoint)
            _endpoints.append(new_ep)
        return self.create_from_endpoints(_endpoints)

    @classmethod
    def create_from_endpoints(cls, endpoints: List[EsiEndpoint], **kwargs):
        """Create stub from endpoints."""
        return cls(testdata=None, endpoints=endpoints, **kwargs)



"""Utilities related to URLs."""

import re
from typing import Optional
from urllib.parse import urljoin

from django.conf import settings
from django.templatetags.static import static
from django.urls import reverse


# old: get_absolute_url
def reverse_absolute(viewname: str, args: Optional[list] = None) -> str:
    """Return absolute URL for a view name.

    Similar to Django's ``reverse()``, but returns an absolute URL.
    """
    return urljoin(site_absolute_url(), reverse(viewname, args=args))


# TODO: Only enable for alliance auth
# old: get_site_base_url
def site_absolute_url() -> str:
    """Return absolute URL for this Alliance Auth site."""
    try:
        match = re.match(r"(.+)\/sso\/callback", settings.ESI_SSO_CALLBACK_URL)
        if match:
            return match.group(1)
    except AttributeError:
        pass

    return ""


def static_file_absolute_url(file_path: str) -> str:
    """Return absolute URL to a static file.

    Args:
        file_path: relative path to a static file
    """
    return urljoin(site_absolute_url(), static(file_path))



"""Utilities related to Alliance Auth."""

# pylint: disable=unused-import

from functools import partial
from typing import Optional

from redis import Redis

from django.contrib.auth.models import Permission, User
from django.core.cache import caches

from allianceauth.notifications import notify
from allianceauth.views import NightModeRedirectView

from ._app_settings import APP_UTILS_NOTIFY_THROTTLED_TIMEOUT
from .django import users_with_permission
from .helpers import throttle
from .testing import create_fake_user  # noqa: F401

try:
    import django_redis
except ImportError:
    django_redis = None


def notify_admins(message: str, title: str, level: str = "info") -> None:
    """Send notification to all admins.

    Args:
        message: Message text
        title: Message title
        level: Notification level of the message.
    """
    try:
        perm = Permission.objects.get(codename="logging_notifications")
    except Permission.DoesNotExist:
        users = User.objects.filter(is_superuser=True)
    else:
        users = users_with_permission(perm)
    for user in users:
        notify(user, title=title, message=message, level=level)


def notify_admins_throttled(
    message_id: str,
    message: str,
    title: str,
    level: str = "info",
    timeout: Optional[int] = None,
):
    """Send notification to all admins, but limits the frequency
    for sending messages with the same message ID, e.g. to once per day.

    If this function is called during a timeout the notification will simply be ignored.

    Args:
        message_id: ID representing this message
        message: Message text
        title: Message title
        level: Notification level of the message.
        timeout: Time between each notification, e.g. 86400 = once per day.\
            When not provided uses system default,\
            which is 86400 and can also be set via this Django setting:\
            APP_UTILS_NOTIFY_THROTTLED_TIMEOUT
    """
    if not timeout:
        timeout = APP_UTILS_NOTIFY_THROTTLED_TIMEOUT
    throttle(
        func=partial(notify_admins, message, title, level),
        context_id=message_id,
        timeout=timeout,
    )


def notify_throttled(
    message_id: str,
    user: User,
    title: str,
    message: str,
    level: str = "info",
    timeout: Optional[int] = None,
):
    """Send notification to user, but limits the frequency
    for sending messages with the same message ID, e.g. to once per day.

    If this function is called during a timeout the notification will simply be ignored.

    Args:
        message_id: ID representing this message
        title: Message title
        message: Message text
        level: Notification level of the message.
        timeout: Time between each notification, e.g. 86400 = once per day.\
            When not provided uses system default,\
            which is 86400 and can also be set via this Django setting:\
            APP_UTILS_NOTIFY_THROTTLED_TIMEOUT
    """
    if not timeout:
        timeout = APP_UTILS_NOTIFY_THROTTLED_TIMEOUT
    throttle(
        func=partial(notify, user, title, message, level),
        context_id=message_id,
        timeout=timeout,
    )


def is_night_mode(request) -> bool:
    """Returns True if the current user session is in night mode, else False"""
    return NightModeRedirectView.night_mode_state(request)


def get_redis_client() -> Redis:
    """Return configured redis client used for Django caching in Alliance Auth."""
    try:
        return django_redis.get_redis_connection("default")  # type: ignore
    except AttributeError:
        default_cache = caches["default"]
        return default_cache.get_master_client()  # type: ignore



"""Utilities for supporting Django views."""

from enum import Enum
from http import HTTPStatus
from typing import List, Optional

from django.http import HttpResponse, JsonResponse
from django.utils.functional import lazy
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

DEFAULT_ICON_SIZE = 32
format_html_lazy = lazy(format_html, str)


class HttpResponseNoContent(HttpResponse):
    """Special HTTP response with no content, just headers.

    The content operations are ignored.
    """

    status_code = HTTPStatus.NO_CONTENT

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # although we don't define a content-type, base class sets a
        # default one -- remove it, we're not returning content
        if hasattr(self, "headers") and "content-type" in self.headers:
            del self.headers["content-type"]
        else:
            raise ValueError("No header found")  # this should never be called


class JSONResponseMixin:
    """A mixin that can be used to render a JSON response for a class based view."""

    def render_to_json_response(self, context, **response_kwargs):
        """Return a JSON response, transforming 'context' to make the payload."""
        return JsonResponse(self.get_data(context), safe=False, **response_kwargs)

    def get_data(self, context):
        """Return an object that will be serialized as JSON by json.dumps()."""
        return context


class BootstrapStyle(str, Enum):
    """A Bootstrap context style names, e.g. for labels.

    DEPRECATED: This class works for Bootstrap 3 only, which is no longer supported by AA.
    """

    DANGER = "danger"  #:
    DEFAULT = "default"  #:
    INFO = "info"  #:
    PRIMARY = "primary"  #:
    SUCCESS = "success"  #:
    WARNING = "warning"  #:

    def __str__(self) -> str:
        return self.value


class BootstrapStyleBS5(str, Enum):
    """A Bootstrap context style name, e.g. for labels"""

    DANGER = "danger"  #:
    DEFAULT = "secondary"  #:
    INFO = "info"  #:
    PRIMARY = "primary"  #:
    SUCCESS = "success"  #:
    WARNING = "warning"  #:

    def __str__(self) -> str:
        return self.value


# old: add_bs_label_html
def bootstrap_label_html(text: str, label: str = "default") -> str:
    """Return HTML for a Bootstrap label."""
    return format_html('<span class="label label-{}">{}</span>', label, text)


# old: create_bs_glyph_html
def bootstrap_glyph_html(glyph_name: str) -> str:
    """Return HTML for a Bootstrap 3 glyph."""
    return format_html(
        '<span class="glyphicon glyphicon-{}"></span>', glyph_name.lower()
    )


# old: create_bs_glyph_2_html
def bootstrap_glyph_2_html(glyph_name, tooltip_text=None, color="initial") -> str:
    """Return HTML for a Bootstrap 3 glyph HTML. Can also have a tool tip and color."""
    if tooltip_text:
        tooltip_html = mark_safe(
            'aria-hidden="true" data-toggle="tooltip" data-placement="top" '
            f'title="{tooltip_text}"'
        )
    else:
        tooltip_html = ""
    return format_html(
        '<span class="glyphicon glyphicon-{}" style="color:{};"{}></span>',
        glyph_name.lower(),
        color,
        tooltip_html,
    )


def bootstrap_icon_plus_name_html(
    icon_url,
    name,
    size: int = DEFAULT_ICON_SIZE,
    avatar: bool = False,
    url: Optional[str] = None,
    text: Optional[str] = None,
) -> str:
    """Returns HTML to display an icon next to a name. Can also be a link.

    DEPRECATED. To be replaced by bootstrap_icon_plus_text_html.
    """
    name_html = link_html(url, name, new_window=False) if url else name
    if text:
        name_html = format_html("{}&nbsp;{}", name_html, text)

    return format_html(
        "{}&nbsp;&nbsp;&nbsp;{}",
        image_html(
            icon_url, classes=["ra-avatar", "img-circle"] if avatar else [], size=size
        ),
        name_html,
    )


def bootstrap_icon_plus_text_html(
    icon_url: str,
    text: str,
    size: int = DEFAULT_ICON_SIZE,
    avatar: bool = False,
    url: Optional[str] = None,
    suffix: Optional[str] = None,
) -> str:
    """Return HTML for a widget consisting of an icon with a text or link:

    `<Icon> <Text/Link> [<Suffix>]`

    Adds the class ``icon-plus-text`` to all images to enable styling.

    Args:
        - icon_url: URL to an icon file, supported by <img>, e.g. GIF, PNG
        - text: Text to be shown next to icon. Often a name.
        - size: Size of the icon in pixel
        - avatar: When enabled the icon will get Bootstrap 3 classes for avatars
        - url: A provided URL will transform the text into a link
        - suffix: Additional text to be shown after the main text.
            Will not be part of a link.
    """
    my_css_class = "icon-plus-text"
    if url:
        text_html = link_html(url, text, new_window=False, classes=[my_css_class])
    else:
        text_html = format_html('<span class="{}">{}</span>', my_css_class, text)

    if suffix:
        text_html = format_html("{} {}", text_html, suffix)

    my_image_html = image_html(
        icon_url, classes=["img-circle"] if avatar else [], size=size
    )
    return format_html("{} {}", my_image_html, text_html)


# old: create_bs_button_html
def bootstrap_link_button_html(
    url: str, glyph_name: str, button_type: str, disabled: bool = False
) -> str:
    """Return HTML for a button with Bootstrap 3 glyphs."""
    return format_html(
        '<a href="{}" class="btn btn-{}"{}>{}</a>',
        url,
        button_type,
        mark_safe(' disabled="disabled"') if disabled else "",
        bootstrap_glyph_html(glyph_name),
    )


# old: create_fa_button_html
def fontawesome_link_button_html(
    url: str,
    fa_code: str,
    button_type: str,
    tooltip: Optional[str] = None,
    disabled: bool = False,
) -> str:
    """Return HTML for a button with fontawesome symbols."""
    return format_html(
        '<a href="{}" class="btn btn-{}"{}>{}{}</a>',
        url,
        button_type,
        mark_safe(f' title="{tooltip}"') if tooltip else "",
        mark_safe(' disabled="disabled"') if disabled else "",
        mark_safe(f'<i class="{fa_code}"></i>'),
    )


def fontawesome_modal_button_html(
    modal_id: str,
    fa_code: str,
    ajax_url: str = "",
    tooltip: str = "",
    style=BootstrapStyle.DEFAULT,
) -> str:
    """Return HTML for a modal button with fontawesome symbols.

    DEPRECATED: Only works with Bootstrap3, which is no longer supported by AA

    Args:
        modal_id: DOM ID of modal to invoke
        fa_code: fontawesome code, e.g. "fas fa-moon"
        ajax_url: URL to invoke via AJAX for loading modal content
        tooltip: text to appear as tooltip
        style: Bootstrap context style for the button
    """
    return format_html(
        '<button type="button" '
        'class="btn btn-{}" '
        'data-toggle="modal" '
        'data-target="#{}" '
        "{}"
        "{}>"
        '<i class="{}"></i>'
        "</button>",
        BootstrapStyle(style),
        modal_id,
        mark_safe(f'title="{tooltip}" ') if tooltip else "",
        mark_safe(f'data-ajax_url="{ajax_url}" ') if ajax_url else "",
        fa_code,
    )


def humanize_value(value: float, precision: int = 2) -> str:
    """Return a value in a human readable and abbreviated form
    e.g. ``1234678`` -> ``1.23m``.
    """
    value = float(value)
    for exponent, identifier in [(12, "t"), (9, "b"), (6, "m"), (3, "k")]:
        if value >= pow(10, exponent):
            return f"{value / pow(10, exponent):,.{precision}f}{identifier}"

    return f"{value:,.{precision}f}"


def image_html(
    src: str, classes: Optional[List[str]] = None, size: Optional[int] = None
) -> str:
    """Return HTML for an image with optional classes and size."""
    classes_html = _classes_html(classes) if classes else ""
    size_html = (
        format_html(' width="{}" height="{}"', int(size), int(size)) if size else ""
    )
    return format_html('<img{}{} src="{}">', classes_html, size_html, src)


# old: create_link_html
def link_html(
    url: str, label: str, new_window: bool = True, classes: Optional[List[str]] = None
) -> str:
    """Return HTML for a link.

    Args:
        - url: URL
        - label: Text describing the link
        - new_window: When enabled the link will open a new window
        - classes: Classes to be added to the a tag
    """
    classes_html = _classes_html(classes) if classes else ""
    target_html = mark_safe(' target="_blank"') if new_window else ""
    return format_html('<a{} href="{}"{}>{}</a>', classes_html, url, target_html, label)


def _classes_html(classes: List[str]) -> str:
    classes_html = format_html(' class="{}"', (" ".join(classes)))
    return classes_html


# old: add_no_wrap_html
def no_wrap_html(text: str) -> str:
    """Add no-wrap HTML to text with Bootstrap 3."""
    return format_html('<span class="text-nowrap">{}</span>', text)


def yesno_str(value: bool) -> str:
    """Return yes/no for boolean as string and with localization."""
    return _("yes") if value is True else _("no")


def yesnonone_str(value: Optional[bool]) -> str:
    """Return yes/no/none for boolean as string and with localization."""
    if value is True:
        return _("yes")
    if value is False:
        return _("no")
    return ""



"""Database related functionality."""


class TableSizeMixin:
    """Add a table size functionality to a Django Manager."""

    def db_table_size(self) -> int:
        """Calculate the database table size in bytes.

        This method works for MySQL/MariaDB databases only.
        """
        from django.db import connection

        is_mysql = connection.settings_dict.get("ENGINE") == "django.db.backends.mysql"
        if not is_mysql:
            raise RuntimeError("This method only works for MySQL like databases.")
        database_name = connection.settings_dict["NAME"]
        table_name = self.model._meta.db_table
        with connection.cursor() as cursor:
            cursor.execute(
                """
                SELECT
                    table_name AS `Table`,
                    round(data_length + index_length, 2)
                FROM information_schema.TABLES
                WHERE table_schema = %s
                    AND table_name = %s;
                """,
                [database_name, table_name],
            )
            row = cursor.fetchone()
        try:
            return int(row[1])
        except (TypeError, IndexError, ValueError):
            raise ValueError(
                f"Failed to calculate table size for {database_name}.{table_name}"
            ) from None



"""Custom test runners for Django."""

import statistics
from time import time
from typing import List, NamedTuple
from unittest.runner import TextTestResult, TextTestRunner

from django.test.runner import DiscoverRunner

SLOWEST_TESTS_TOP_COUNT = 10


class _TestDurationResult(NamedTuple):
    name: str  # Name of the test
    duration: float  # duration in seconds


_results: List[_TestDurationResult] = []


class _TimedTextTestResult(TextTestResult):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.clocks = {}

    def startTest(self, test):
        self.clocks[test] = time()
        super().startTest(test)
        if self.showAll:
            self.stream.write(self.getDescription(test))
            self.stream.write(" ... ")
            self.stream.flush()

    def addSuccess(self, test):
        super().addSuccess(test)
        duration = time() - self.clocks[test]
        _results.append(_TestDurationResult(name=str(test), duration=duration))
        if self.showAll:
            self.stream.writeln(f"{duration:.2f}s")
        elif self.dots:
            self.stream.write(".")
            self.stream.flush()


class _TimedTextTestRunner(TextTestRunner):
    resultclass = _TimedTextTestResult

    def run(self, test):
        result = super().run(test)
        self._output_slowest_tests()
        return result

    def _output_slowest_tests(self):
        """Write list of slowest tests to output."""
        self.stream.writeln()
        self.stream.writeln(f"Top {SLOWEST_TESTS_TOP_COUNT} slowest tests:")
        _results.sort(reverse=True, key=lambda o: o.duration)
        for obj in _results[:SLOWEST_TESTS_TOP_COUNT]:
            self.stream.writeln(f"{obj.duration:>8.2f}s: {obj.name}")
        values = [obj.duration for obj in _results]
        average_duration = statistics.mean(values)
        tests_count = len(values)
        self.stream.writeln(
            f"Average test duration of {tests_count} tests was {average_duration:.2f}s."
        )
        self.stream.writeln()
        self.stream.flush()


class TimedTestRunner(DiscoverRunner):
    """Test runner which adds duration measurements to each tests (when verbose)
    and shows list of slowest tests at the end.

    To use in tests define via the ``TEST_RUNNER`` setting
    or as ``--testrunner`` parameter for ``test``, e.g.:

    .. code-block:: python

        TEST_RUNNER = "app_utils.testrunners.TimedTestRunner"


    .. code-block:: bash

        python manage.py test -v 2 --testrunner app_utils.testrunners.TimedTestRunner

    """

    test_runner = _TimedTextTestRunner



"""Django settings related utilities."""

import logging
from typing import Any, Optional

from django.conf import settings

logger = logging.getLogger(__name__)


def clean_setting(
    name: str,
    default_value: Any,
    min_value: Optional[int] = None,
    max_value: Optional[int] = None,
    required_type: Optional[type] = None,
    choices: Optional[list] = None,
) -> Any:
    """Clean a setting from Django settings.

    Will use default_value if setting is not defined.
    Will use minimum or maximum value if respective boundary is exceeded.

    Args:
        default_value: value to use if setting is not defined
        min_value: minimum allowed value (0 assumed for int)
        max_value: maximum value value
        required_type: Mandatory if `default_value` is `None`,
        otherwise derived from default_value

    Returns:
        cleaned value for setting

    This function is designed to be used in a dedicated module like ``app_settings.py``
    as layer between the actual settings and all other modules.
    ``app_settings.py`` will import and clean all settings and all other modules are supposed to import the settings it.

    Example for app_settings:

    .. code-block:: python

        from app_utils.app_settings import clean_setting

        EXAMPLE_SETTING = clean_setting("EXAMPLE_SETTING", 10)
    """
    if default_value is None and not required_type:
        raise ValueError("You must specify a required_type for None defaults")

    if not required_type:
        required_type_2 = type(default_value)
    else:
        required_type_2 = required_type

    if not isinstance(required_type_2, type):
        raise TypeError("required_type must be a type when defined")

    if min_value is None and issubclass(required_type_2, int):
        min_value = 0

    if issubclass(required_type_2, int) and default_value is not None:
        if min_value is not None and default_value < min_value:
            raise ValueError("default_value can not be below min_value")
        if max_value is not None and default_value > max_value:
            raise ValueError("default_value can not be above max_value")

    if not hasattr(settings, name):
        cleaned_value = default_value
    else:
        dirty_value = getattr(settings, name)
        if dirty_value is None or (
            isinstance(dirty_value, required_type_2)
            and (min_value is None or dirty_value >= min_value)
            and (max_value is None or dirty_value <= max_value)
            and (choices is None or dirty_value in choices)
        ):
            cleaned_value = dirty_value
        elif (
            isinstance(dirty_value, required_type_2)
            and min_value is not None
            and dirty_value < min_value
        ):
            logger.warning(
                "You setting for %s it not valid. Please correct it. "
                "Using minimum value for now: %s",
                name,
                min_value,
            )
            cleaned_value = min_value
        elif (
            isinstance(dirty_value, required_type_2)
            and max_value is not None
            and dirty_value > max_value
        ):
            logger.warning(
                "You setting for %s it not valid. Please correct it. "
                "Using maximum value for now: %s",
                name,
                max_value,
            )
            cleaned_value = max_value
        else:
            logger.warning(
                "You setting for %s it not valid. Please correct it. "
                "Using default for now: %s",
                name,
                default_value,
            )
            cleaned_value = default_value
    return cleaned_value



"""Settings configuration for app_utils."""

from .app_settings import clean_setting

APP_UTILS_NOTIFY_THROTTLED_TIMEOUT = clean_setting(
    "APP_UTILS_NOTIFY_THROTTLED_TIMEOUT", 86400
)
"""Timeout for throttled notifications in seconds."""

APPUTILS_ESI_ERROR_LIMIT_THRESHOLD = clean_setting(
    "APPUTILS_ESI_ERROR_LIMIT_THRESHOLD", 25
)
"""ESI error limit remain threshold.

The number of remaining errors is counted down from 100 as errors occur.
Because multiple tasks may request the value simultaneously and get the same response,
the threshold must be above 0 to prevent the API from shutting down with a 420 error.
"""

APPUTILS_ESI_DAILY_DOWNTIME_START = clean_setting("APPUTILS_ESI_DOWNTIME_START", 11.0)
"""Start time of daily downtime in UTC hours.

esi.fetch_esi_status() will report ESI as offline during this time.
"""

APPUTILS_ESI_DAILY_DOWNTIME_END = clean_setting("APPUTILS_ESI_DOWNTIME_END", 11.25)
"""End time of daily downtime in UTC hours.

esi.fetch_esi_status() will report ESI as offline during this time.
"""



"""General purpose helpers."""

import hashlib
import os
import random
import string
from typing import Any, Callable, Optional

from django.core.cache import cache


def chunks(lst, size):
    """Yield successive sized chunks from lst."""
    for i in range(0, len(lst), size):
        yield lst[i : i + size]


def default_if_none(value, default):
    """Return default if value is None."""
    if value is None:
        return default
    return value


# old: get_swagger_spec_path
def swagger_spec_path() -> str:
    """returns the path to the current esi swagger spec file"""
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "swagger.json")


def random_string(char_count: int) -> str:
    """returns a random string of given length"""
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(char_count)
    )


class AttrDict(dict):
    """Enhanced dict that allows property access to its keys.

    Example:

    .. code-block:: python

        >> my_dict = AttrDict({"color": "red", "size": "medium"})
        >> my_dict["color"]
        "red"
        >> my_dict.color
        "red"

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


def humanize_number(value, magnitude: Optional[str] = None, precision: int = 1) -> str:
    """Return the value in humanized format, e.g. `1234` becomes `1.2k`

    Args:
        magnitude: fix the magnitude to format the number, e.g. `"b"`
        precision: number of digits to round for
    """
    value = float(value)
    power_map = {"t": 12, "b": 9, "m": 6, "k": 3, "": 0}
    if magnitude not in power_map:
        if value >= 10**12:
            magnitude = "t"
        elif value >= 10**9:
            magnitude = "b"
        elif value >= 10**6:
            magnitude = "m"
        elif value >= 10**3:
            magnitude = "k"
        else:
            magnitude = ""
    return f"{value / 10 ** power_map[magnitude]:,.{precision}f}{magnitude}"


def throttle(func: Callable, context_id: str, timeout: Optional[int]) -> Any:
    """Call a function, but limit repeated calls with a timeout, e.g. once per day.

    When a repeated call falls within the timeout the call will simply be ignored.

    Args:
        func: the function to be called
        context_id: a string representing the context for applying the throttle,\
            e.g. a combination of feature name and user ID
        timeout: timeout in seconds between each repeated call of the function

    Returns:
        Return cached value of called function func

    """
    hashed_id = hashlib.md5(str(context_id).encode("utf-8")).hexdigest()
    key = f"APP_UTILS_THROTTLED_{hashed_id}"
    return cache.get_or_set(key, func, timeout)
